'''
    The disassemblers are only responsible for the formatting of tokens and are
    _not_ responsible for any parsing or tokenizing.
'''

# Standard library import statements
from __future__ import print_function

import re

# c8d's import statements
import chip8.parser
import chip8.lexer
import chip8.settings


def fmt_addr(addr):
    '''
        Formats an address label by removing the '0x' portsion and capitalizing
        it. In assembler, labels appear like 'A12B:' and not like '0xA12B:'.
    '''
    return re.sub('0x', '', addr).upper()


def fmt_comment(instruction):
    '''
        Formats the end of the line comments, the comment includes the original
        opcode that produced the assembler. This can be useful to trace back to
        the original bytes within the file that caused a bug etc.
    '''
    stripped = re.sub('0x', '', instruction).upper()

    return '; opcode: 0x{inst}'.format(inst=stripped.zfill(4))


class C8Disassembler(object):
    '''
        The disassembler takes the tokens produced by the lexer and displays
        the tokens in a format that makes more sense to a human.
    '''

    def __init__(self, file_path):
        self.parser = chip8.parser.C8Parser(file_path)
        self.lexer = chip8.lexer.C8Lexer(self.parser)

    def disassemble(self):
        '''
            Begin the process by telling the lexer to tokenize the file.
        '''
        self.lexer.tokenize()

        return self

    def print_info(self):
        '''
            Print information on c8d and the file being disassembled.
        '''
        build = chip8.settings.build()

        c8d_info = '; c8d version {major_v}.{minor_v}, {author_info}'.format(
            major_v=build['major_version'],
            minor_v=build['minor_version'],
            author_info='{author} ({author_email})'.format(
                author=build['author'],
                author_email=build['author_email']
            )
        )

        file_info = '; file: {filename} ({filesize} bytes)'.format(
            filename=self.parser.filename(),
            filesize=self.parser.total_bytes
        )

        print(c8d_info)
        print(file_info)
        print('')

    def print_tokens(self):
        '''
            Print the game's disassembled instructions.
        '''
        for token in self.lexer.tokens:
            self.print_token(token)

    def print_token(self, token):
        '''
            Given a token, print it in a humanly readable format.
        '''

        # Token format is always (offset, instruction, ...)
        offset = token[0]
        instruction = token[1].ljust(4)

        if offset + 0x200 in self.lexer.labels:
            label = fmt_addr(hex(offset + 0x200)).rjust(4)
            print('{label}:'.format(label=label))

        if len(token) == 3:
            comment = fmt_comment(token[2].rjust(4))

            print('\t{0}{1}{2}'.format(instruction, ''.ljust(22), comment))
        elif len(token) == 4:
            op1 = token[2]
            comment = fmt_comment(token[3])

            print(
                '\t{0} {1},{2}{3}'.format(
                    instruction, op1, ''.ljust(20 - len(op1)), comment
                )
            )
        elif len(token) == 5:
            op1 = token[2]
            op2 = token[3].rjust(3)
            comment = fmt_comment(token[4])

            print(
                '\t{0} {1}, {2}{3}{4}'.format(
                    instruction,
                    op1,
                    op2,
                    ''.ljust(19 - len(op1) - len(op2)),
                    comment
                )
            )
        elif len(token) == 6:
            op1 = token[2]
            op2 = token[3].rjust(3)
            op3 = token[4].rjust(2)
            comment = fmt_comment(token[5])

            print(
                '\t{0} {1}, {2}, {3}{4}{5}'.format(
                    instruction,
                    op1,
                    op2,
                    op3,
                    ''.ljust(17 - len(op1) - len(op2) - len(op3)),
                    comment
                )
            )

    def run(self):
        '''
            Read the tokens produced by the lexer and print them in a more
            friendly format.
        '''

        self.print_info()
        self.print_tokens()
