'''
    Lexers are responsible for tokenizing the machine instructions into a
    series of tuples which can be processed by the disassembler.
'''

import re


# pylint: disable=invalid-name

def v(reg):
    '''
        Takes a register number (0x0 -> 0xF) and formats it into the VX format.
    '''

    hex_val = re.sub('0x', '', hex(reg))

    return 'V{register}'.format(register=hex_val.upper())


def imm(value):
    '''
        Takes a value and formats it as #X.
    '''

    hex_val = re.sub('0x', '', hex(value))

    return '#{upper}'.format(upper=hex_val.upper().zfill(2))


def label(value):
    '''
        Take an address and formats it in the label format, which does not
        include the '0x' portion of the hex value.
    '''

    hex_val = re.sub('0x', '', hex(value))

    return '{upper}'.format(upper=hex_val.upper().zfill(2))


def x(opcode):
    '''
        Extracts the X portion of the opcode (0x0XYN).
    '''
    return (opcode & 0x0F00) >> 8


def y(opcode):
    '''
        Extracts the Y portion of the opcode (0x0XYN).
    '''
    return (opcode & 0x00F0) >> 4


def n(opcode):
    '''
        Extracts the N portion of the opcode (0x0XYN).
    '''
    return opcode & 0x000F


def nn(opcode):
    '''
        Extracts the NN portion of the opcode (0x0XNN).
    '''
    return opcode & 0x00FF


def nnn(opcode):
    '''
        Extracts the NNN portion of the opcode (0xXNNN).
    '''
    return opcode & 0x0FFF

# pylint: enable=invalid-name


class C8Lexer(object):
    '''
        The lexer takes the raw machine instructions and tokenizes them into
        a set of tuples. Each tuple contains the instruction number, a string
        representing the instruction and any operands for the instruction.
    '''

    def __init__(self, parser):
        self.parser = parser

        self.labels = []
        self.tokens = []

    def tokenize(self):
        '''
            Tokenizes each instruction within the game's data.
        '''
        self.parser.parse()

        for inst_p in range(0, self.parser.total_bytes, 2):
            self.decode_instruction(self.parser.data, inst_p)

    def decode_instruction(self, data, inst_p):
        # pylint: disable=too-many-branches,too-many-statements
        '''
            Decodes a single instruction into a token tuple and inserts it into
            the list of tokens.
        '''

        # Combine the current byte with the next byte to form the opcode, each
        # opcode is 16-bits wide.
        opcode = (data[inst_p] << 8) | (data[inst_p + 1])

        # Note: This could be converted into a jump table if performance was
        # an issue, but Chip-8 games are small enough, this shouldn't ever be
        # an issue.
        if opcode & 0xF000 == 0x0:
            if (opcode & 0x00FF) == 0x0E0:
                self.tokens.append((inst_p, 'clr', hex(opcode)))
            elif (opcode & 0x00FF) == 0xEE:
                self.tokens.append((inst_p, 'rts', hex(opcode)))
        elif opcode & 0xF000 == 0x1000:
            self.tokens.append((inst_p, 'jp', label(nnn(opcode)), hex(opcode)))
            self.labels.append(nnn(opcode))
        elif opcode & 0xF000 == 0x2000:
            self.tokens.append(
                (inst_p, 'call', label(nnn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0x3000:
            self.tokens.append(
                (inst_p, 'se', v(x(opcode)), imm(nn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0x4000:
            self.tokens.append(
                (inst_p, 'sne', v(x(opcode)), imm(nn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0x5000:
            self.tokens.append(
                (inst_p, 'se', v(x(opcode)), v(y(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0x6000:
            self.tokens.append(
                (inst_p, 'ld', v(x(opcode)), imm(nn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0x7000:
            self.tokens.append(
                (inst_p, 'add', v(x(opcode)), imm(nn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0x8000:
            if opcode & 0x000F == 0x0:
                self.tokens.append(
                    (inst_p, 'ld', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x01:
                self.tokens.append(
                    (inst_p, 'or', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x02:
                self.tokens.append(
                    (inst_p, 'and', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x03:
                self.tokens.append(
                    (inst_p, 'xor', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x04:
                self.tokens.append(
                    (inst_p, 'add', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x05:
                self.tokens.append(
                    (inst_p, 'sub', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x06:
                self.tokens.append(
                    (inst_p, 'shr', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x07:
                self.tokens.append(
                    (inst_p, 'subn', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
            elif opcode & 0x000F == 0x0E:
                self.tokens.append(
                    (inst_p, 'shl', v(x(opcode)), v(y(opcode)), hex(opcode))
                )
        elif opcode & 0xF000 == 0x9000:
            self.tokens.append(
                (inst_p, 'sne', v(x(opcode)), v(y(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0xA000:
            self.tokens.append(
                (inst_p, 'ld', ' I', label(nnn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0xB000:
            self.tokens.append(
                (inst_p, 'jp', 'V0', label(nnn(opcode)), hex(opcode))
            )
            self.labels.append(nnn(opcode))
        elif opcode & 0xF000 == 0xC000:
            self.tokens.append(
                (inst_p, 'rnd', v(x(opcode)), imm(nn(opcode)), hex(opcode))
            )
        elif opcode & 0xF000 == 0xD000:
            self.tokens.append(
                (
                    inst_p,
                    'drw',
                    v(x(opcode)),
                    v(y(opcode)),
                    label(n(opcode)),
                    hex(opcode)
                )
            )
        elif opcode & 0xF000 == 0xE000:
            if opcode & 0x00FF == 0x9E:
                self.tokens.append(
                    (inst_p, 'skp', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0xA1:
                self.tokens.append(
                    (inst_p, 'sknp', v(x(opcode)), hex(opcode))
                )
        elif opcode & 0xF000 == 0xF000:
            if opcode & 0x00FF == 0x07:
                self.tokens.append(
                    (inst_p, 'ld', v(x(opcode)), 'DT', hex(opcode))
                )
            elif opcode & 0x00FF == 0x0A:
                self.tokens.append(
                    (inst_p, 'ld', v(x(opcode)), 'K', hex(opcode))
                )
            elif opcode & 0x00FF == 0x15:
                self.tokens.append(
                    (inst_p, 'ld', 'DT', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0x18:
                self.tokens.append(
                    (inst_p, 'ld', 'ST', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0x1E:
                self.tokens.append(
                    (inst_p, 'add', ' I', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0x29:
                self.tokens.append(
                    (inst_p, 'ld', ' F', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0x33:
                self.tokens.append(
                    (inst_p, 'ld', ' B', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0x55:
                self.tokens.append(
                    (inst_p, 'ld', '[I]', v(x(opcode)), hex(opcode))
                )
            elif opcode & 0x00FF == 0x65:
                self.tokens.append(
                    (inst_p, 'ld', v(x(opcode)), '[I]', hex(opcode))
                )
